#!/usr/bin/env python
# weili@jimmy.harvard.edu
'''
input: binary version of Affy bpmap file, import Affy BPMAP SDK
Two stage Redundance Removal as described in HMMTiling Bioinformatics Paper
Output: Text version of bpmap file,   *.bpmap.NonRe.txt
'''

import sys, random,re
sys.path.append( re.sub('[^\/]+$', 'SDK',sys.argv[0]))
import BPMAP                #SDK
from numarray import *
DISTANCE = 1000   # 1kb as short-range repeat probe measurements

class Mybpmap(object):
    '''store data in bpmapfile, self.seq['chr21'] = seqObj'''

    def __init__(self, bpmapname):
        '''read the bpmap file first to get the seq(chromosome) information, but del the data to save memory
        you have to use the Read() to get the real data
        '''
        bpmap = BPMAP.CBPMAPFileData()
        bpmap.SetFileName(bpmapname)
        if not bpmap.Exists():
            bpmap.SetLibPath('.')
        if not bpmap.Exists():
            print >> sys.stderr, 'Cannot open', bpmapname
            sys.exit(0)
        print >> sys.stderr, 'Reading ', bpmapname
        bpmap.Read()

        self.seq = {}
        for iseq in range(bpmap.GetNumberSequences() ):
            seq = BPMAP.CGDACSequenceItem()
            bpmap.GetSequenceItem(iseq, seq)
            self.seq[seq.GetName()] = seq
        self.bpmap = bpmap

        #bpmapname + .NR.bpmap.txt as the text format non-redundant bpmap file
        outfile = re.compile('bpmap', re.I).sub('bpmap.NR.txt', bpmapname)
        if outfile == bpmapname:
            outfile = bpmapname + '.txt'
        self.out = open(outfile, 'w')
        print >> self.out, "X\tY\tSeq\tPos\tProbe"

        #to save memory
        self.bpmap.Close()

    def Read(self):
        print >> sys.stderr, 'Reading ', self.bpmap.GetFileName()
        self.bpmap.Read()

    def Close(self):
        self.bpmap.Close()

    def GetProbe(self, chr):
        '''get data structure, [chipid,  probeid,  PMX, PMY, pos] from one chromosome (seq)'''
        nhit = self.seq[chr].GetNumberHits()
        hit = BPMAP.GDACSequenceHitItemType()

        data = zeros( (nhit, 5))
        for ihit in range( nhit ):
            self.seq[chr].GetHitItem(ihit, hit, 1)
            data[ihit] = [0, ihit, hit.PMX, hit.PMY,  hit.Position]
        return data

    def save(self, chr, id):
        '''save to plain-text format'''
        hit = BPMAP.GDACSequenceHitItemType()
        self.seq[chr].GetHitItem(id, hit, 1)
        print >> self.out, "%d\t%d\t%s\t%d\t%s" % \
        (hit.PMX, hit.PMY, chr, hit.Position, self.bpmap.GetStr(hit.PMProbe))

    def savebinary(self):
        '''save to the regular binary format, reserved'''
        pass

def main(args):
    bps = [] # list for bpmap files
    for x in args:
        bps.append(Mybpmap(x))

    chrs= [] # list for chromosomes in all bpmap files
    for bp in bps:
        for chr in bp.seq.keys():
            if not chrs.count(chr):
                chrs.append(chr)
    chrs.sort()

    # remove redundancy for each chromosome
    for chr in chrs:
        probes = []
        print >> sys.stderr, 'processing ', chr
        for i, bp in enumerate(bps):
            if bp.seq.has_key(chr):
                bp.Read()
                tmp = bp.GetProbe(chr)
                tmp[:,0] = i
                probes.append(tmp)
        probe = concatenate(probes)

        # sorting probe according to position column at 5th,
        #data structure, [chipid,  probeid,  PMX, PMY, pos] from one chromosome (seq)'''
        probe = take(probe, probe[:,4].argsort() )

        print >> sys.stderr, "removing redundance stage 1..."
        pos = {}
        for i,ipro in enumerate(probe):
            m = pos.get(( ipro[0], ipro[2], ipro[3] ))                   #chipid, PMX, PMY
            if m:                                   #find the same probe
                if ipro[4] <= m + DISTANCE:         # within 1kb short-range
                    print >> sys.stderr, 'deleting  ', ipro
                    probe[i] = -1                   #delet probe[i]
                    continue
            pos[ (ipro[0], ipro[2], ipro[3]) ] = ipro[4]                #no duplication found, continue index

        #remove probes marked as -1
        probe = compress(probe[:,4] != -1, probe )

        print >> sys.stderr, "removing redundance stage 2..."
        repeatlines = []
        for i in range(len(probe) -1 ):
            repeatlines.append(i)
            if probe[i+1][4] == probe[i][4] :        #probes are on the same pos
                print >> sys.stderr, 'may deleting  ', probe[i+1]
            else:
                if len(repeatlines) >1 :                #find previous duplication
                    tmp = random.choice(repeatlines)    #choose one entry
                    tmpentry = probe[tmp].copy()
                    probe[repeatlines] = -1
                    probe[tmp] = tmpentry
                repeatlines = []
        else:                                       #handle the last line
            if len(repeatlines) >0 :                #if last lines are duplication, only pick the last line
                probe[repeatlines] = -1

        #remove probes marked as -1
        probe = compress(probe[:,4] != -1, probe )

        print >> sys.stderr, "writing to disk ... "
        for ipro in probe:
            bps[ipro[0]].save(chr, ipro[1])
        #close all bps in this chromosome to save memory
        for i, bp in enumerate(bps):
            if bp.seq.has_key(chr):
                bp.Close()


#read program options
if __name__ == '__main__':
    import sys

    if len(sys.argv) <2:
        print "Usage: %prog bpmapfile1 bpmapfile2 ..."
        print '''
        This program will make non-redundant text-format bpmap files from
        the original Affymetrix binary bpmap files. It uses two-stage Redundancy
        Removal to ensure unique probe measurement within 1 kb. It will add NR.txt
        suffix in the non-redundant bpmap files'''
        sys.exit(0)

    main(sys.argv[1:])