#!/usr/bin/env python
# weili@jimmy.harvard.edu

'''
Cel2Matrix.py
Input: Normalized Cel files and BPMap file
Output: Matrix files, one chromosome (sequence) each, Or One big file contains all chrs
'''
import sys, numarray, re, time
sys.path.append( re.sub('[^\/]+$', 'SDK',sys.argv[0]))
import BPMAPtxt, CEL             #SDK

class CelList(object):
    '''self.bpmap = {'bpmap1':[cel1, cel2,...,celN], 'bpmap2':[cel1,...]}'''
    def __init__(self, filename):
        self.bpmap={}
        filehandle = open(filename)
        header = filehandle.readline().split()
        if not (header[0] == 'cel_file' and header[1] == 'bpmap_file'):
            print >> sys.stderr, "Wrong fromat in your cel-file list, please check ", filename
            sys.exit()
        for x in filehandle:
            x = x.split()
            if not self.bpmap.has_key( x[1] ):
                self.bpmap[ x[1] ] = []
            self.bpmap[ x[1] ].append(x[0])

        #sorting and checking
        temp = self.bpmap.keys()
        temp.sort()
        ncel = len( self.bpmap[temp[0]] )
        for key in temp:
            self.bpmap[key].sort()
            if len(self.bpmap[key] ) != ncel:
                print >> sys.stderr, len(self.bpmap[key] )+ " in "+ key + " is not the same as " + ncel + " in " + temp[0]
                sys.exit()

class Mycel(object):
    def __init__(self, args, path):
        import numarray, sys, CEL
        self.cels = []
        self.nintensity = len(args)
        for i in range(len(args)):
            print >> sys.stderr, 'reading ', args[i]
            cel = CEL.CCELFileData()
            cel.SetFileName(args[i])
            cel.SetDataPath(path)
            if not cel.Exists():
                print >> sys.stderr, "cannot open cel file ",args[i]
                sys.exit(0)
            cel.Read()
            self.cels.append(cel)

    def getintensity(self, x, y, outlier):
        entry = CEL.CELFileEntryType()
        intensity = numarray.zeros(self.nintensity, numarray.Float32)

        for i, cel in enumerate(self.cels):
            cel.GetEntry(x,y,entry)

            if outlier and cel.IsOutlier(x, y):             #mark outlier probes as -999999
                intensity[i] = -999999
            else:
                intensity[i] = entry.Intensity
        return intensity


    def getintensityPM_MM(self, x, y, outlier):
        ''' either pm or mm is an outlier, the pm-mm value is marked as -999999, MMY = PMY +1'''
        entry = CEL.CELFileEntryType()
        intensity = numarray.zeros(self.nintensity, numarray.Float32)

        for i, cel in enumerate(self.cels):
            cel.GetEntry(x,y,entry)
            pm = entry.Intensity
            cel.GetEntry(x, y+1, entry)
            mm = entry.Intensity

            if outlier and ( cel.IsOutlier(x, y) or cel.IsOutlier(x,y+1)):          #mark outlier probes as -999999
                intensity[i] = -999999
            else:
                intensity[i] = pm - mm
        return intensity


def main(options):
    cellist = CelList(options.celfiles)
    bpfilename = cellist.bpmap.keys()
    bpfilename.sort()

    bps = [] # list for bpmap files
    for x in bpfilename:
        bps.append(BPMAPtxt.Bpmaptxt(x, options.dir))

    chrs= [] # list for chromosomes in all bpmap files
    for bp in bps:
        for chr in bp.seq.keys():
            if not chrs.count(chr):
                chrs.append(chr)
    chrs.sort()

    for chr in chrs:
        #output file:
        outfile = open(options.outfile+'.'+chr+'.matrix', 'w')
        print >> outfile, "#"+chr
        print >> outfile, "#", time.ctime(),"Running parametres:  " , " ".join(sys.argv)

        probes = []
        print >> sys.stderr, 'processing ', chr
        for i, bp in enumerate(bps):
            if bp.seq.has_key(chr): #bpmap contains this chromosome
                bp.Read()
                tmp = bp.GetProbe(chr)
                tmp[:,0] = i
                probes.append(tmp)

                #reading all cel files with this bpmap file
                bps[i].cels = Mycel( cellist.bpmap[bp.filename], options.dir )  # Mycel Obj

                #print cel header
                print >> outfile,"ChipID\tPos\t"+ "\t".join( [x.split('.')[0] for x in cellist.bpmap[bp.filename]] )

        # sorting probe according to position column
        #data structure, [chipid,  PMX, PMY, pos] from one chromosome (seq)
        probe = numarray.concatenate(probes)
        probe = numarray.take(probe, probe[:,3].argsort() )

        print >> sys.stderr, "Writing ", chr
        for (chipid, pmx, pmy, pos) in probe:
            if options.method == 1:          #PM-MM
                intensity = bps[chipid].cels.getintensityPM_MM(pmx, pmy, options.outlier)
                intensity = ["%.1f" % x for x in intensity ]
                print >> outfile, str(chipid) + "\t" +str(pos) + "\t" + "\t".join(intensity)
            if options.method == 2:         #PM and MM respectively for Plier.Tiling
                pm =  ["%.1f" % x for x in bps[chipid].cels.getintensity(pmx, pmy, options.outlier) ]
                mm =  ["%.1f" % x for x in bps[chipid].cels.getintensity(pmx, pmy +1, options.outlier) ]
                print >> outfile,str(chipid) + "\t" +str(pos) + "\t" + "\t".join(pm)
                print >> outfile, str(chipid) + "\t" +str(pos) + "\t" + "\t".join(mm)

        #close all bps in this chromosome to save memory
        for i, bp in enumerate(bps):
            if bp.seq.has_key(chr):
                bp.Close()
                del bp.cels

if __name__ == '__main__':

    import optparse
    parser = optparse.OptionParser(usage = "%prog  options cel1 cel2 ..", version = "%prog 06052005",\
                          formatter = optparse.TitledHelpFormatter(),\
    description = '''This program will take a set of normalized cel files and the text-version non-redundant bpmap files to create a matrix file. One matrix file per chromosome. The cel files and bpmap files should be in the same folder. In the matrix file, the first two columns are ChipID and Position, the rightwards columns are different array samples.''')

    parser.add_option("-o", dest="outfile",
                      help="output file name, will add .chr.matrix suffix ")

    parser.add_option("-m", dest="method",type="int",
                  help='''-m [1,2,3,4,5]; method 1:  PM-MM;  method 2: PM and MM respectively
                ''')

    parser.add_option("--cel-files", dest="celfiles", metavar="FILE",
                      help = '''provide a list of cel files; It's a text file and should
                      have a cel_file and bpmap_file columns,                        ''')

    parser.add_option("--dir",dest="dir",
                      help = "Location of your cel and bpmap files")

    parser.add_option("--outlier", dest='outlier', action = 'store_true', default= False,
                      help = 'mark all the outlier probes ([OUTLIERS] section in cel files) as -999999')

    (options, args) = parser.parse_args()

    if len(sys.argv) <2:
        parser.print_help()
        sys.exit()
    if not (options.method == 1 or options.method ==2) :
        parser.error("Please specify your method, 1 or 2")

    main(options)
