/////////////////////////////////////////////////////////////////
//
// Copyright (C) 2004 Affymetrix, Inc.
//
// This library is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published
// by the Free Software Foundation; either version 2.1 of the License,
// or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
// for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this library; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
//
/////////////////////////////////////////////////////////////////

#include "AB_CELFile.h"
#include "CELFileData.h"
#include <assert.h>
#include <string>

using namespace affxbruce;

///////////////////////////////////////////////////////////////////////////////
///  public constructor  CELFile
///  \brief Default constructor
///
///  @return void	
///////////////////////////////////////////////////////////////////////////////
CELFile::CELFile()
{
	m_pCelFileData = new affxcel::CCELFileData();
}

///////////////////////////////////////////////////////////////////////////////
///  public destructor  ~CELFile
///  \brief Destructor
///
///  @return void	
///////////////////////////////////////////////////////////////////////////////
CELFile::~CELFile()
{
	if (m_pCelFileData != NULL)
		delete m_pCelFileData;
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetError
///  \brief Call affxcel::CCELFileData::SetError() to set error string
///
///  @param  str char * 	Error string
///  @return void	
///
///  @see affxcel::CCELFileData::SetError
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetError(char* str)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetError(str);
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetError
///  \brief Call affxcel::CCELFileData::GetError() to retrieve error string
///
///  @return const char *	string that includes error generated by function execution
///
///  @see affxcel::CCELFileData::GetError
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetError() const
{
	assert(m_pCelFileData != NULL);
	char* szError = NULL;
	CopyString(szError, m_pCelFileData->GetError().c_str());
	return szError;
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetDataPath
///  \brief Call affxcel::CCELFileData::SetDataPath() to set data path 
///
///  @param  str const char * 	File path that does not contain file name
///  @return void	
///
///  @see affxcel::CCELFileData::SetDataPath
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetDataPath(const char* str)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetDataPath(str);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetLibPath
///  \brief Call affxcel::CCELFileData::SetLibPath() to set lib path
///
///  @param  str const char * 	Lib path
///  @return void	
///
///  @see affxcel::CCELFileData::SetLibPath
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetLibPath(const char* str)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetLibPath(str);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetFileName
///  \brief Call affxcel::CCELFileData::SetFileName() to set file name
///
///  @param  str const char * 	File name without path
///  @return void	
///
///  @see affxcel::CCELFileData::SetFileName
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetFileName(const char* str)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetFileName(str);
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetFileName
///  \brief Call affxcel::CCELFileData::GetFileName() to retrieve file name
///
///  @return const char *	File name without path
///
///  @see affxcel::CCELFileData::GetFileName
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetFileName() const
{
	assert(m_pCelFileData != NULL);
	char* szFileName = NULL;
	CopyString(szFileName, m_pCelFileData->GetFileName().c_str());
	return szFileName;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetFilePath
///  \brief Call affxcel::CCELFileData::GetFilePath() to retrieve file path
///
///  @return const char *	Full file path including data path and file name
///
///  @see affxcel::CCELFileData::GetFilePath
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetFilePath() const
{
	assert(m_pCelFileData != NULL);
	char* szFilePath = NULL;
	CopyString(szFilePath, m_pCelFileData->GetFilePath().c_str());
	return szFilePath;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetHeaderKey
///  \brief Call affxcel::CCELFileData::GetHeaderKey() to retrieve header value specified by key
///
///  @param  key char * 	Header key
///  @return const char *	Header value
///
///  @see affxcel::CCELFileData::GetHeaderKey
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetHeaderKey(char* key) const
{
	assert(m_pCelFileData != NULL);
	char* szHeaderValue = NULL;
	CopyString(szHeaderValue, m_pCelFileData->GetHeaderKey(key).c_str());
	return szHeaderValue;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetVersion
///  \brief Call affxcel::CCELFileData::GetVersion() to retrieve file version number
///
///  @return int	Version number
///
///  @see affxcel::CCELFileData::GetVersion
///////////////////////////////////////////////////////////////////////////////
int CELFile::GetVersion() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetVersion();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetCols
///  \brief Call affxcel::CCELFileData::GetCols() to retrieve number of columns in array
///
///  @return int	Number of columns
///
///  @see affxcel::CCELFileData::GetCols
///////////////////////////////////////////////////////////////////////////////
int CELFile::GetCols() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetCols();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetRows
///  \brief Call affxcel::CCELFileData::GetRows() to retrieve number of rows in array
///
///  @return int	Number of rows
///
///  @see affxcel::CCELFileData::GetRows
///////////////////////////////////////////////////////////////////////////////
int CELFile::GetRows() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetRows();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetNumCells
///  \brief Call affxcel::CCELFileData::GetNumCells() to retrieve number of cells in array
///
///  @return int	Number of cells
///
///  @see affxcel::CCELFileData::GetNumCells
///////////////////////////////////////////////////////////////////////////////
int CELFile::GetNumCells() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetNumCells();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetHeader
///  \brief Call affxcel::CCELFileData::GetHeader() to retrieve header information
///
///  @return const char *	Header info concatenated in string
///
///  @see affxcel::CCELFileData::GetHeader
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetHeader() const
{
	assert(m_pCelFileData != NULL);
	char* szHeader = NULL;
	CopyString(szHeader, m_pCelFileData->GetHeader().c_str());
	return szHeader;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetAlg
///  \brief Call affxcel::CCELFileData::GetAlg() to retrieve algorithm name
///
///  @return const char *	Algorithm name
///
///  @see affxcel::CCELFileData::GetAlg
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetAlg() const
{
	assert(m_pCelFileData != NULL);
	char* szAlg = NULL;
	CopyString(szAlg, m_pCelFileData->GetAlg().c_str());
	return szAlg;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetParams
///  \brief Call affxcel::CCELFileData::GetParams() to retrieve algorithm parameters
///
///  @return const char *	Algorithm parameters
///
///  @see affxcel::CCELFileData::GetParams
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetParams() const
{
	assert(m_pCelFileData != NULL);
	char* szParams = NULL;
	CopyString(szParams, m_pCelFileData->GetParams().c_str());
	return szParams;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetDatHeader
///  \brief Call affxcel::CCELFileData::GetDatHeader() to retrieve DAT header
///
///  @return const char *	DAT header
///
///  @see affxcel::CCELFileData::GetDatHeader
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetDatHeader() const
{
	assert(m_pCelFileData != NULL);
	char* szDatHeader = NULL;
	CopyString(szDatHeader, m_pCelFileData->GetDatHeader().c_str());
	return szDatHeader;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetChipType
///  \brief Call affxcel::CCELFileData::GetChipType() to retrieve chip type
///
///  @return const char *	Chip type
///
///  @see affxcel::CCELFileData::GetChipType
///////////////////////////////////////////////////////////////////////////////
const char* CELFile::GetChipType() const
{
	assert(m_pCelFileData != NULL);
	char* szChipType = NULL;
	CopyString(szChipType, m_pCelFileData->GetChipType().c_str());
	return szChipType;
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetCellMargin
///  \brief Call affxcel::CCELFileData::GetCellMargin() to retrieve cell margin
///
///  @return int	Cell margin
///
///  @see affxcel::CCELFileData::GetCellMargin
///////////////////////////////////////////////////////////////////////////////
int CELFile::GetCellMargin() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetCellMargin();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetNumOutliers
///  \brief Call affxcel::CCELFileData::GetNumOutliers() to retrieve number of outliers
///
///  @return unsigned long	Number of outliers
///
///  @see affxcel::CCELFileData::GetNumOutliers
///////////////////////////////////////////////////////////////////////////////
unsigned long CELFile::GetNumOutliers() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetNumOutliers();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetNumMasked
///  \brief Call affxcel::CCELFileData::GetNumMasked() to retrieve number of masked cells
///
///  @return unsigned long	Number of masked cells
///
///  @see affxcel::CCELFileData::GetNumMasked
///////////////////////////////////////////////////////////////////////////////
unsigned long CELFile::GetNumMasked() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetNumMasked();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetCorners
///  \brief Call affxcel::CCELFileData::GetCorners() to retrieve grid coordinates
///
///  @return affxcel::GridCoordinatesType	Grid coordinates
///
///  @see affxcel::CCELFileData::GetCorners
///////////////////////////////////////////////////////////////////////////////
affxcel::GridCoordinatesType CELFile::GetCorners() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetCorners();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetMean
///  \brief Call affxcel::CCELFileData::GetMean() to retrieve mean intensity of specified cell
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @return float	Mean intensity
///
///  @see affxcel::CCELFileData::GetMean
///////////////////////////////////////////////////////////////////////////////
float CELFile::GetMean(int x, int y) const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetMean(x, y);
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetStdev
///  \brief Call affxcel::CCELFileData::GetStdev() to retrieve standard deviation intensity of specified cell
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate
///  @return float	Standard deviation intensity
///
///  @see affxcel::CCELFileData::GetStdev
///////////////////////////////////////////////////////////////////////////////
float CELFile::GetStdev(int x, int y) const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetStdev(x, y);
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetPixels
///  \brief Call affxcel::CCELFileData::GetPixels() to retrieve number of pixels of specified cell
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @return short	Number of pixels
///
///  @see affxcel::CCELFileData::GetPixels
///////////////////////////////////////////////////////////////////////////////
short CELFile::GetPixels(int x, int y) const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetPixels(x, y);
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  IsMasked
///  \brief Call affxcel::CCELFileData::IsMasked() to determine if the specified cell is masked
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @return int	1 for masked cell; 0 for unmasked cell
///
///  @see affxcel::CCELFileData::IsMasked
///////////////////////////////////////////////////////////////////////////////
int CELFile::IsMasked(int x, int y) const
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->IsMasked(x, y)? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  IsOutlier
///  \brief Call affxcel::CCELFileData::IsOutlier() to determine if the specified cell is outlier
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @return int	1 for outlier; 0 for not an outlier
///
///  @see affxcel::CCELFileData::IsOutlier
///////////////////////////////////////////////////////////////////////////////
int CELFile::IsOutlier(int x, int y) const
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->IsOutlier(x, y)? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  Close
///  \brief Call affxcel::CCELFileData::Close() to close memory mapped file and clear variables
///
///  @return void
///
///  @see affxcel::CCELFileData::Close
///////////////////////////////////////////////////////////////////////////////
void CELFile::Close()
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->Close();
}

///////////////////////////////////////////////////////////////////////////////
///  public constant  GetFileSize
///  \brief Call affxcel::CCELFileData::GetFileSize() to retrieve file size
///
///  @return long	File size
///
///  @see affxcel::CCELFileData::GetFileSize
///////////////////////////////////////////////////////////////////////////////
long CELFile::GetFileSize() const
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetFileSize();
}

///////////////////////////////////////////////////////////////////////////////
///  public  Exists
///  \brief Call affxcel::CCELFileData::Exists() to determine if file exists
///
///  @return int	1 for file exists; 0 for file not exists
///
///  @see affxcel::CCELFileData::Exists
///////////////////////////////////////////////////////////////////////////////
int CELFile::Exists()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->Exists()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  ReadHeader
///  \brief Call affxcel::CCELFileData::ReadHeader() to read header section only of the CEL file
///
///  @return int	1 for success; 0 for fail
///
///  @see affxcel::CCELFileData::ReadHeader
///////////////////////////////////////////////////////////////////////////////
int CELFile::ReadHeader()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->ReadHeader()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  Read
///  \brief Call affxcel::CCELFileData::Read() to read CEL file
///
///  @return int	1 for success; 0 for fail
///
///  @see affxcel::CCELFileData::Read
///////////////////////////////////////////////////////////////////////////////
int CELFile::Read()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->Read()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  IsXDACompatibleFile
///  \brief Call affxcel::CCELFileData::IsXDACompatibleFile() to determine if CEL file is in XDA format
///
///  @return int	1 for XDA; 0 for others
///
///  @see affxcel::CCELFileData::IsXDACompatibleFile
///////////////////////////////////////////////////////////////////////////////
int CELFile::IsXDACompatibleFile()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->IsXDACompatibleFile()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  IsTranscriptomeBcelFile
///  \brief Call affxcel::CCELFileData::IsTranscriptomeBcelFile() to determine if CEL file is in bcel format
///
///  @return int	1 for bcel; 0 for others
///
///  @see affxcel::CCELFileData::IsTranscriptomeBcelFile
///////////////////////////////////////////////////////////////////////////////
int CELFile::IsTranscriptomeBcelFile()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->IsTranscriptomeBcelFile()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  IsCompactCelFile
///  \brief Call affxcel::CCELFileData::IsCompactCelFile() to determine if CEL file is in compact format
///
///  @return int	1 for ccel; 0 for others
///
///  @see affxcel::CCELFileData::IsCompactCelFile
///////////////////////////////////////////////////////////////////////////////
int CELFile::IsCompactCelFile()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->IsCompactCelFile()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  Clear
///  \brief Call affxcel::CCELFileData::Clear() to reset values of all variables
///
///  @return void
///
///  @see affxcel::CCELFileData::Clear
///////////////////////////////////////////////////////////////////////////////
void CELFile::Clear()
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->Clear();
}

///////////////////////////////////////////////////////////////////////////////
///  public  GetFileFormat
///  \brief Call affxcel::CCELFileData::GetFileFormat() to retrieve CEL file format
///
///  @return int	1 for text; 2 for xda; 3 for bcel; 4 for ccel
///
///  @see affxcel::CCELFileData::GetFileFormat
///////////////////////////////////////////////////////////////////////////////
int CELFile::GetFileFormat()
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetFileFormat();
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetFileFormat
///  \brief Call affxcel::CCELFileData::SetFileFormat() to set CEL file format
///
///  @param  i int  	CEL file format index
///  @return void	
///
///  @see affxcel::CCELFileData::SetFileFormat
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetFileFormat(int i)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetFileFormat(i);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetAlgorithmName
///  \brief Call affxcel::CCELFileData::SetAlgorithmName() to set algorithm name
///
///  @param  str const char * 	Algorithm name
///  @return void	
///
///  @see affxcel::CCELFileData::SetAlgorithmName
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetAlgorithmName(const char *str)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetAlgorithmName(str);
}

///////////////////////////////////////////////////////////////////////////////
///  public  AddAlgorithmParameter
///  \brief Call affxcel::CCELFileData::AddAlgorithmParameter() to add algorithm parameter
///
///  @param  tag const char * 	Algorithm parameter tag
///  @param  value const char * 	Algorithm parameter value
///  @return void	
///
///  @see affxcel::CCELFileData::AddAlgorithmParameter
///////////////////////////////////////////////////////////////////////////////
void CELFile::AddAlgorithmParameter(const char *tag, const char *value)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->AddAlgorithmParameter(tag, value);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetDimensions
///  \brief Call affxcel::CCELFileData::SetDimensions() to set array dimensions
///
///  @param  rows int  	Number of rows in array
///  @param  cols int  	Number of columns in array
///  @return void	
///
///  @see affxcel::CCELFileData::SetDimensions
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetDimensions(int rows, int cols)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetDimensions(rows, cols);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetChipType
///  \brief Call affxcel::CCELFileData::SetChipType() to set chip type
///
///  @param  str const char * 	Chip type
///  @return void	
///
///  @see affxcel::CCELFileData::SetChipType
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetChipType(const char *str)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetChipType(str);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetMargin
///  \brief Call affxcel::CCELFileData::SetMargin() to set cell margin
///
///  @param  margin int  	Cell margin
///  @return void	
///
///  @see affxcel::CCELFileData::SetMargin
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetMargin(int margin)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetMargin(margin);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetMean
///  \brief Call affxcel::CCELFileData::SetMean() to set mean intensity of specified cell
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @param  mean float  	Mean intensity
///  @return void	
///
///  @see affxcel::CCELFileData::SetMean
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetMean(int x, int y, float mean)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetMean(x, y, mean);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetStdev
///  \brief Call affxcel::CCELFileData::SetStdev() to set standard deviation intensity of specified cell
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @param  stdev float  	Standard deviation intensity
///  @return void	
///
///  @see affxcel::CCELFileData::SetStdev
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetStdev(int x, int y, float stdev)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetStdev(x, y, stdev);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetPixels
///  \brief Call affxcel::CCELFileData::SetPixels() to set number of pixels of specified cell
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @param  pixels short  	Number of pixels
///  @return void	
///
///  @see affxcel::CCELFileData::SetPixels
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetPixels(int x, int y, short pixels)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetPixels(x, y, pixels);
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetMasked
///  \brief Call affxcel::CCELFileData::SetMasked() to set if the specified cell is masked
///
///  @param  x int  	X coodinate of cell
///  @param  y int  	Y coodinate of cell
///  @param  masked int  	1 for masked; 0 for unmasked
///  @return void	
///
///  @see affxcel::CCELFileData::SetMasked
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetMasked(int x, int y, int masked)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetMasked(x, y, ((masked == 1)? true : false));
}

///////////////////////////////////////////////////////////////////////////////
///  public  SetOutlier
///  \brief Call affxcel::CCELFileData::SetOutlier() to set if the specified cell is an outlier
///
///  @param  x int  	X coordinate of cell
///  @param  y int  	Y coordinate of cell
///  @param  outlier int  	1 for outlier; 0 for not an outlier
///  @return void	
///
///  @see affxcel::CCELFileData::SetOutlier
///////////////////////////////////////////////////////////////////////////////
void CELFile::SetOutlier(int x, int y, int outlier)
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->SetOutlier(x, y, ((outlier == 1)? true : false));
}

///////////////////////////////////////////////////////////////////////////////
///  public  WriteTextCel
///  \brief Call affxcel::CCELFileData::WriteTextCel() to write CEL file in text format
///
///  @return int	1 for success; 0 for fail
///
///  @see affxcel::CCELFileData::WriteTextCel
///////////////////////////////////////////////////////////////////////////////
int CELFile::WriteTextCel()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->WriteTextCel()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  WriteXDABCel
///  \brief Call affxcel::CCELFileData::WriteXDABCel() to write CEL file in XDA format
///
///  @return int	1 for success; 0 for fail
///
///  @see affxcel::CCELFileData::WriteXDABCel
///////////////////////////////////////////////////////////////////////////////
int CELFile::WriteXDABCel()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->WriteXDABCel()? 1 : 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  WriteTranscriptomeBCel
///  \brief Call affxcel::CCELFileData::WriteTranscriptomeBCel() to write CEL file in bcel format
///
///  @return int	1 for success; 0 for fail
///
///  @see affxcel::CCELFileData::WriteTranscriptomeBCel
///////////////////////////////////////////////////////////////////////////////
int CELFile::WriteTranscriptomeBCel()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->WriteTranscriptomeBCel()?  1: 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  WriteCompactBCel
///  \brief Call affxcel::CCELFileData::WriteCompactBCel() to write CEL file in ccel format
///
///  @return int	1 for success; 0 for fail
///
///  @see affxcel::CCELFileData::WriteCompactBCel
///////////////////////////////////////////////////////////////////////////////
int CELFile::WriteCompactBCel()
{
	assert(m_pCelFileData != NULL);
	return (m_pCelFileData->WriteCompactBCel()?  1: 0);
}

///////////////////////////////////////////////////////////////////////////////
///  public  GetThisPtr
///  \brief Call affxcel::CCELFileData::GetThisPtr() to retrieve the current pointer of CEL file object
///
///  @return affxcel::CCELFileData *	Point to CEL file object
///
///  @see affxcel::CCELFileData::GetThisPtr
///////////////////////////////////////////////////////////////////////////////
affxcel::CCELFileData *CELFile::GetThisPtr()
{
	assert(m_pCelFileData != NULL);
	return m_pCelFileData->GetThisPtr();
}

///////////////////////////////////////////////////////////////////////////////
///  public  EnsureNotMmapped
///  \brief Call affxcel::CCELFileData::EnsureNotMmapped() to put CEL file data in memory instead of using memory mapping
///
///  @return void	
///
///  @see affxcel::CCELFileData::EnsureNotMmapped
///////////////////////////////////////////////////////////////////////////////
void CELFile::EnsureNotMmapped()
{
	assert(m_pCelFileData != NULL);
	m_pCelFileData->EnsureNotMmapped();
}

///////////////////////////////////////////////////////////////////////////////
///  public  TEXT_CEL
///  \brief Call affxcel::CCELFileData::TEXT_CEL to retrieve text format index
///
///  @return int	Text format index
///
///  @see affxcel::CCELFileData::TEXT_CEL
///////////////////////////////////////////////////////////////////////////////
int CELFile::TEXT_CEL()
{
	return affxcel::CCELFileData::TEXT_CEL;
}

///////////////////////////////////////////////////////////////////////////////
///  public  XDA_BCEL
///  \brief Call affxcel::CCELFileData::XDA_BCEL to retrieve XDA format index
///
///  @return int	XDA format index
///
///  @see affxcel::CCELFileData::XDA_BCEL
///////////////////////////////////////////////////////////////////////////////
int CELFile::XDA_BCEL()
{
	return affxcel::CCELFileData::XDA_BCEL;
}

///////////////////////////////////////////////////////////////////////////////
///  public  TRANSCRIPTOME_BCEL
///  \brief Call affxcel::CCELFileData::TRANSCRIPTOME_BCEL to retrieve bcel format index
///
///  @return int	Bcel format index
///
///  @see affxcel::CCELFileData::TRANSCRIPTOME_BCEL
///////////////////////////////////////////////////////////////////////////////
int CELFile::TRANSCRIPTOME_BCEL()
{
	return affxcel::CCELFileData::TRANSCRIPTOME_BCEL;
}

///////////////////////////////////////////////////////////////////////////////
///  public  COMPACT_BCEL
///  \brief Call affxcel::CCELFileData::COMPACT_BCEL to retrieve ccel format index
///
///  @return int	Ccel format index
///
///  @see affxcel::CCELFileData::COMPACT_BCEL
///////////////////////////////////////////////////////////////////////////////
int CELFile::COMPACT_BCEL()
{
	return affxcel::CCELFileData::COMPACT_BCEL;
}

///////////////////////////////////////////////////////////////////////////////
///  private constant  CopyString
///  \brief Copy one string to another
///
///  @param  pDest char *& 	Destinated string
///  @param  pSource const char * 	Source string
///  @return void	
///////////////////////////////////////////////////////////////////////////////
void CELFile::CopyString(char*& pDest, const char* pSource) const
{
	int iLen = (int) strlen(pSource);
	if (pDest != NULL)
	{
		delete [] pDest;
		pDest = NULL;
	}
	pDest = new char[iLen + 1];
	memcpy(pDest, pSource, iLen * sizeof(char));
	pDest[iLen] = '\0';
}

///////////////////////////////////////////////////////////////////////////////
///  private constant  AppendString
///  \brief Append source string to destination
///
///  @param  pDest char *& 	Destinated string
///  @param  pSource const char * 	Source string
///  @return void	
///////////////////////////////////////////////////////////////////////////////
void CELFile::AppendString(char*& pDest, const char* pSource) const
{
	int iSourceLen = (int) strlen(pSource);
	if (pDest == NULL)
		pDest = new char[iSourceLen + 1];
	else
	{
		int iDestLen = (int) strlen(pDest);
		char* pTemp = new char[iDestLen + iSourceLen + 1];
		memset(pTemp, 0, (iDestLen + iSourceLen + 1) * sizeof(char));
		memcpy(pTemp, pDest, iDestLen * sizeof(char));
		delete [] pDest;
		pDest = pTemp;
	}
	pDest = strcat(pDest, pSource);
}

//////////////////////////////

namespace std { // ns

///////////////////////////////////////////////////////////////////////////////
///  global CelFilePtrVector_pushref
///  \brief Add pointer to CEL file object into vector
///
///  @param  celptrvec std::vector<CCELFileData *> & 	Vector of pointers to CEL file object
///  @param  cel CCELFileData * 	Pointer to CEL file object to be added to vector
///  @return void	
///////////////////////////////////////////////////////////////////////////////
void CelFilePtrVector_pushref(std::vector<CCELFileData *>&celptrvec,CCELFileData *cel) {
  celptrvec.push_back(cel);
  //printf("size=%d\n",celptrvec.size());
}

///////////////////////////////////////////////////////////////////////////////
///  global CelFilePtrVector_getref
///  \brief Retrieve pointer to CEL file object from vector
///
///  @param  celptrvec std::vector<CCELFileData *> & 	Vector of pointers to CEL file object
///  @param  i unsigned int		Index of pointer to CEL file object to be retrieved
///  @return  cel CCELFileData * 	Pointer to CEL file object
///////////////////////////////////////////////////////////////////////////////
CCELFileData *CelFilePtrVector_getref(std::vector<CCELFileData *>&celptrvec,unsigned int i) {
  if ((0<=i)&&(i<celptrvec.size())) {
    return celptrvec[i];
  } else {
    return NULL;
  }
}

///////////////////////////////////////////////////////////////////////////////
///  global CelFilePtrVector_print
///  \brief Print vector content in terms of full file name to STDOUT
///
///  @param  celptrvec std::vector<CCELFileData *> & 	Vector of pointers to CEL file object
///  @return  void
///////////////////////////////////////////////////////////////////////////////
void CelFilePtrVector_print(std::vector<affxcel::CCELFileData *>&celptrvec) {
  int j=0;
  printf("CelFilePtrVector == \n");

  for (std::vector<affxcel::CCELFileData*>::iterator i=celptrvec.begin(); i!=celptrvec.end();i++) {
    cout << j++ << " FILE=" << (*i)->GetFilePath() << "\n";
  }
}
} // ns
